import logging

from . import AuxiliaryCoordinate, Constructs, core, mixin
from .decorators import (
    _display_or_return,
    _inplace_enabled,
    _inplace_enabled_define_and_cleanup,
)

logger = logging.getLogger(__name__)


class Domain(
    mixin.FieldDomain,
    mixin.NetCDFVariable,
    mixin.NetCDFGeometry,
    mixin.NetCDFGlobalAttributes,
    mixin.NetCDFGroupAttributes,
    mixin.NetCDFComponents,
    mixin.NetCDFUnreferenced,
    mixin.Properties,
    mixin.Files,
    core.Domain,
):
    """A domain construct of the CF data model.

    The domain represents a set of discrete "locations" in what
    generally would be a multi-dimensional space, either in the real
    world or in a model's simulated world. The data array elements of
    a field construct correspond to individual location of a domain.

    The domain construct is defined collectively by the following
    constructs of the CF data model: domain axis, dimension
    coordinate, auxiliary coordinate, cell measure, coordinate
    reference, and domain ancillary constructs; as well as properties
    to describe the domain.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF global attributes}}

    {{netCDF group attributes}}

    {{netCDF geometry group}}

    Some components exist within multiple constructs, but when written
    to a netCDF dataset the netCDF names associated with such
    components will be arbitrarily taken from one of them. The netCDF
    variable, dimension and sample dimension names and group
    structures for such components may be set or removed consistently
    across all such components with the `nc_del_component_variable`,
    `nc_set_component_variable`, `nc_set_component_variable_groups`,
    `nc_clear_component_variable_groups`,
    `nc_del_component_dimension`, `nc_set_component_dimension`,
    `nc_set_component_dimension_groups`,
    `nc_clear_component_dimension_groups`,
    `nc_del_component_sample_dimension`,
    `nc_set_component_sample_dimension`,
    `nc_set_component_sample_dimension_groups`, and
    `nc_clear_component_sample_dimension_groups` methods.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes.

        .. versionadded:: (cfdm) 1.7.0

        """
        instance = super().__new__(cls)
        instance._AuxiliaryCoordinate = AuxiliaryCoordinate
        instance._Constructs = Constructs
        return instance

    def __init__(
        self, properties=None, source=None, copy=True, _use_data=True
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                   ``properties={'long_name': 'Domain for model'}``

            {{init source: optional}}

                A new domain may also be instantiated with the
                `fromconstructs` class method.

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        if source is not None:
            try:
                mesh_id = source.get_mesh_id(None)
            except AttributeError:
                pass
            else:
                if mesh_id is not None:
                    self.set_mesh_id(mesh_id)

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

        self._set_dataset_compliance(self.dataset_compliance(), copy=True)

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        shape = sorted(
            [
                domain_axis.get_size(None)
                for domain_axis in self.domain_axes(todict=True).values()
            ]
        )
        shape = str(shape)
        shape = shape[1:-1]

        return f"<{self.__class__.__name__}: {self._one_line_description()}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """

        def _print_item(self, cid, variable, axes):
            """Private function called by __str__."""
            x = [variable.identity(default=f"key%{cid}")]

            if variable.has_data():
                shape = [axis_names[axis] for axis in axes]
                data = variable.get_data()
                ndim = data.ndim
                shape = shape[:ndim]
                if len(shape) < ndim:
                    shape.extend([str(n) for n in data.shape[len(shape) :]])
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(",)", ")")
                x.append(shape)
            elif (
                variable.construct_type
                in ("auxiliary_coordinate", "domain_ancillary")
                and variable.has_bounds()
                and variable.bounds.has_data()
            ):
                # Construct has no data but it does have bounds
                shape = [axis_names[axis] for axis in axes]
                shape.extend(
                    [str(n) for n in variable.bounds.data.shape[len(axes) :]]
                )
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(",)", ")")
                x.append(shape)
            elif (
                hasattr(variable, "nc_get_external")
                and variable.nc_get_external()
            ):
                ncvar = variable.nc_get_variable(None)
                if ncvar is not None:
                    x.append(f" (external variable: ncvar%{ncvar})")
                else:
                    x.append(" (external variable)")

            if variable.has_data():
                x.append(f" = {variable.data}")
            elif (
                variable.construct_type
                in ("auxiliary_coordinate", "domain_ancillary")
                and variable.has_bounds()
                and variable.bounds.has_data()
            ):
                # Construct has no data but it does have bounds data
                x.append(f" = {variable.bounds.data}")

            return "".join(x)

        string = []

        axis_names = self._unique_domain_axis_identities()

        construct_data_axes = self.constructs.data_axes()

        x = []
        dimension_coordinates = self.dimension_coordinates(todict=True)
        for axis_cid, axis in sorted(self.domain_axes(todict=True).items()):
            for cid, dim in dimension_coordinates.items():
                if construct_data_axes[cid] == (axis_cid,):
                    name = dim.identity(default=f"key%{0}")
                    y = f"{name}({axis.get_size()})"
                    if y != axis_names[axis_cid]:
                        y = f"{name}({axis_names[axis_cid]})"

                    if dim.has_data():
                        y += f" = {dim.get_data()}"
                    else:
                        y += " = "

                    x.append(y)

        if x:
            x = "\n                : ".join(x)
            string.append(f"Dimension coords: {x}")

        # Auxiliary coordinates
        x = [
            _print_item(self, cid, v, construct_data_axes[cid])
            for cid, v in sorted(
                self.auxiliary_coordinates(todict=True).items()
            )
        ]
        if x:
            x = "\n                : ".join(x)
            string.append(f"Auxiliary coords: {x}")

        # Cell measures
        x = [
            _print_item(self, cid, v, construct_data_axes[cid])
            for cid, v in sorted(self.cell_measures(todict=True).items())
        ]
        if x:
            x = "\n                : ".join(x)
            string.append(f"Cell measures   : {x}")

        # Coordinate references
        x = sorted(
            [
                str(ref)
                for ref in list(
                    self.coordinate_references(todict=True).values()
                )
            ]
        )
        if x:
            x = "\n                : ".join(x)
            string.append(f"Coord references: {x}")

        # Domain ancillary variables
        x = [
            _print_item(self, cid, anc, construct_data_axes[cid])
            for cid, anc in sorted(
                self.domain_ancillaries(todict=True).items()
            )
        ]
        if x:
            x = "\n                : ".join(x)
            string.append(f"Domain ancils   : {x}")

        # Domain topologies
        x = [
            _print_item(self, cid, v, construct_data_axes[cid])
            for cid, v in sorted(self.domain_topologies(todict=True).items())
        ]
        if x:
            x = "\n                : ".join(x)
            string.append(f"Topologies      : {x}")

        # Cell connectivities
        x = [
            _print_item(self, cid, v, construct_data_axes[cid])
            for cid, v in sorted(self.cell_connectivities(todict=True).items())
        ]
        if x:
            x = "\n                : ".join(x)
            string.append(f"Connectivities  : {x}")

        return "\n".join(string)

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    @_display_or_return
    def _dump_axes(self, axis_names, display=True, _level=0):
        """Returns a string description of the field's domain axes.

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

            _level: `int`, optional

        :Returns:

            `str`
                A string containing the description.

        """
        indent1 = "    " * _level

        w = sorted(
            [
                f"{indent1}Domain Axis: {axis_names[axis]}"
                for axis in self.domain_axes(todict=True)
            ]
        )

        return "\n".join(w)

    def _one_line_description(self, axis_names_sizes=None):
        """Return a one-line description of the domain.

        :Returns:

            `str`
                The description.

        """
        if axis_names_sizes is None:
            axis_names_sizes = self._unique_domain_axis_identities()

        axis_names = ", ".join(sorted(axis_names_sizes.values()))

        return f"{self.identity('')}{{{axis_names}}}"

    @_inplace_enabled(default=False)
    def apply_masking(self, inplace=False):
        """Apply masking as defined by the CF conventions.

        Masking is applied to all metadata constructs with data.

        Masking is applied according to any of the following criteria
        that are applicable:

        * where data elements are equal to the value of the
          ``missing_value`` property;

        * where data elements are equal to the value of the
          ``_FillValue`` property;

        * where data elements are strictly less than the value of the
          ``valid_min`` property;

        * where data elements are strictly greater than the value of
          the ``valid_max`` property;

        * where data elements are within the inclusive range specified
          by the two values of ``valid_range`` property.

        If any of the above properties have not been set the no
        masking is applied for that method.

        Elements that are already masked remain so.

        .. note:: If using the `apply_masking` method on a construct
                  that has been read from a dataset with the
                  ``mask=False`` parameter to the `read` function,
                  then the mask defined in the dataset can only be
                  recreated if the ``missing_value``, ``_FillValue``,
                  ``valid_min``, ``valid_max``, and ``valid_range``
                  properties have not been updated.

        .. versionadded:: (cfdm) 1.8.9.0

        .. seealso:: `{{package}}.Data.apply_masking`, `read`, `write`

        :Parameters:

            {{inplace: `bool`, optional}}

        :Returns:

            `Domain` or `None`
                A new domain construct with masked values, or `None`
                if the operation was in-place.

        **Examples**

        >>> d = {{package}}.example_domain(0)
        >>> x = d.construct('longitude')
        >>> x.data[[0, -1]] = {{package}}.masked
        >>> print(x.data.array)
        [-- 67.5 112.5 157.5 202.5 247.5 292.5 --]
        >>> {{package}}.write(d, 'masked.nc')
        >>> no_mask = {{package}}.read('masked.nc', domain=True, mask=False)[0]
        >>> no_mask_x = no_mask.construct('longitude')
        >>> print(no_mask_x.data.array)
        [9.96920997e+36 6.75000000e+01 1.12500000e+02 1.57500000e+02
         2.02500000e+02 2.47500000e+02 2.92500000e+02 9.96920997e+36]
        >>> masked = no_mask.apply_masking()
        >>> masked_x = masked.construct('longitude')
        >>> print(masked_x.data.array)
        [-- 67.5 112.5 157.5 202.5 247.5

        """
        d = _inplace_enabled_define_and_cleanup(self)

        # Apply masking to the metadata constructs
        d._apply_masking_constructs()

        return d

    def climatological_time_axes(self):
        """Return all axes which are climatological time axes.

        This is ascertained by inspecting the values returned by each
        coordinate construct's `is_climatology` method.

        .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `set`
                The keys of the domain axis constructs that are
                climatological time axes.

        **Examples**

        >>> d = {{package}}.example_domain(0)
        >>> d.climatological_time_axes()
        set()

        """
        data_axes = self.constructs.data_axes()

        out = []

        for ckey, c in self.coordinates(todict=True).items():
            if not c.is_climatology():
                continue

            out.extend(data_axes.get(ckey, ()))

        return set(out)

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="domain",
        data_name="data",
        header=True,
        _domain=True,
    ):
        """Return the commands that would create the domain construct.

        **Construct keys**

        The *key* parameter of the output `set_construct` commands is
        utilised in order minimise the number of commands needed to
        implement cross-referencing between constructs (e.g. between a
        coordinate reference construct and coordinate
        constructs). This is usually not necessary when building
        domain constructs, as by default the `set_construct` method
        returns a unique construct key for the construct being set.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `set_construct`,
                     `{{package}}.Data.creation_commands`,
                     `{{package}}.example_domain`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> f = {{package}}.example_domain(0)
        >>> print(d.creation_commands())
        #
        # domain:
        domain = {{package}}.Domain()
        #
        # domain_axis: ncdim%lat
        c = {{package}}.DomainAxis()
        c.set_size(5)
        c.nc_set_dimension('lat')
        domain.set_construct(c, key='domainaxis0', copy=False)
        #
        # domain_axis: ncdim%lon
        c = {{package}}.DomainAxis()
        c.set_size(8)
        c.nc_set_dimension('lon')
        domain.set_construct(c, key='domainaxis1', copy=False)
        #
        # domain_axis:
        c = {{package}}.DomainAxis()
        c.set_size(1)
        domain.set_construct(c, key='domainaxis2', copy=False)
        #
        # dimension_coordinate: latitude
        c = {{package}}.DimensionCoordinate()
        c.set_properties({'units': 'degrees_north', 'standard_name': 'latitude'})
        c.nc_set_variable('lat')
        data = {{package}}.Data([-75.0, -45.0, 0.0, 45.0, 75.0], units='degrees_north', dtype='f8')
        c.set_data(data)
        b = {{package}}.Bounds()
        b.nc_set_variable('lat_bnds')
        data = {{package}}.Data([[-90.0, -60.0], [-60.0, -30.0], [-30.0, 30.0], [30.0, 60.0], [60.0, 90.0]], units='degrees_north', dtype='f8')
        b.set_data(data)
        c.set_bounds(b)
        domain.set_construct(c, axes=('domainaxis0',), key='dimensioncoordinate0', copy=False)
        #
        # dimension_coordinate: longitude
        c = {{package}}.DimensionCoordinate()
        c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
        c.nc_set_variable('lon')
        data = {{package}}.Data([22.5, 67.5, 112.5, 157.5, 202.5, 247.5, 292.5, 337.5], units='degrees_east', dtype='f8')
        c.set_data(data)
        b = {{package}}.Bounds()
        b.nc_set_variable('lon_bnds')
        data = {{package}}.Data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0], [135.0, 180.0], [180.0, 225.0], [225.0, 270.0], [270.0, 315.0], [315.0, 360.0]], units='degrees_east', dtype='f8')
        b.set_data(data)
        c.set_bounds(b)
        domain.set_construct(c, axes=('domainaxis1',), key='dimensioncoordinate1', copy=False)
        #
        # dimension_coordinate: time
        c = {{package}}.DimensionCoordinate()
        c.set_properties({'units': 'days since 2018-12-01', 'standard_name': 'time'})
        c.nc_set_variable('time')
        data = {{package}}.Data([31.0], units='days since 2018-12-01', dtype='f8')
        c.set_data(data)
        domain.set_construct(c, axes=('domainaxis2',), key='dimensioncoordinate2', copy=False)
        >>> print(d.creation_commands(representative_data=True, namespace='',
        ...                           indent=4, header=False))
            domain = Domain()
            c = DomainAxis()
            c.set_size(5)
            c.nc_set_dimension('lat')
            domain.set_construct(c, key='domainaxis0', copy=False)
            c = DomainAxis()
            c.set_size(8)
            c.nc_set_dimension('lon')
            domain.set_construct(c, key='domainaxis1', copy=False)
            c = DomainAxis()
            c.set_size(1)
            domain.set_construct(c, key='domainaxis2', copy=False)
            c = DimensionCoordinate()
            c.set_properties({'units': 'degrees_north', 'standard_name': 'latitude'})
            c.nc_set_variable('lat')
            data = <{{repr}}Data(5): [-75.0, ..., 75.0] degrees_north>  # Representative data
            c.set_data(data)
            b = Bounds()
            b.nc_set_variable('lat_bnds')
            data = <{{repr}}Data(5, 2): [[-90.0, ..., 90.0]] degrees_north>  # Representative data
            b.set_data(data)
            c.set_bounds(b)
            domain.set_construct(c, axes=('domainaxis0',), key='dimensioncoordinate0', copy=False)
            c = DimensionCoordinate()
            c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
            c.nc_set_variable('lon')
            data = <{{repr}}Data(8): [22.5, ..., 337.5] degrees_east>  # Representative data
            c.set_data(data)
            b = Bounds()
            b.nc_set_variable('lon_bnds')
            data = <{{repr}}Data(8, 2): [[0.0, ..., 360.0]] degrees_east>  # Representative data
            b.set_data(data)
            c.set_bounds(b)
            domain.set_construct(c, axes=('domainaxis1',), key='dimensioncoordinate1', copy=False)
            c = DimensionCoordinate()
            c.set_properties({'units': 'days since 2018-12-01', 'standard_name': 'time'})
            c.nc_set_variable('time')
            data = <{{repr}}Data(1): [2019-01-01 00:00:00]>  # Representative data
            c.set_data(data)
            domain.set_construct(c, axes=('domainaxis2',), key='dimensioncoordinate2', copy=False)

        """
        if name in ("b", "c", "mask", "i"):
            raise ValueError(
                f"The 'name' parameter can not have the value {name!r}"
            )

        if name == data_name:
            raise ValueError(
                "The 'name' parameter can not have the same value as "
                f"the 'data_name' parameter: {name!r}"
            )

        namespace0 = namespace
        if namespace is None:
            namespace = self._package() + "."
        elif namespace and not namespace.endswith("."):
            namespace += "."

        if _domain:
            out = super().creation_commands(
                indent=indent,
                namespace=namespace,
                string=False,
                name=name,
                header=header,
            )

            nc_global_attributes = self.nc_global_attributes()
            if nc_global_attributes:
                if header:
                    out.append("#")
                    out.append("# netCDF global attributes")

                out.append(
                    f"{name}.nc_set_global_attributes("
                    f"{nc_global_attributes!r})"
                )
        else:
            out = []

        # Domain axis constructs
        for key, c in self.domain_axes(todict=True).items():
            out.extend(
                c.creation_commands(
                    indent=0,
                    string=False,
                    namespace=namespace0,
                    name="c",
                    header=header,
                )
            )
            out.append(f"{name}.set_construct(c, key={key!r}, copy=False)")

        # Metadata constructs with data
        for key, c in self.constructs.filter_by_type(
            "dimension_coordinate",
            "auxiliary_coordinate",
            "cell_measure",
            "domain_ancillary",
            "domain_topology",
            "cell_connectivity",
        ).items():
            out.extend(
                c.creation_commands(
                    representative_data=representative_data,
                    string=False,
                    indent=0,
                    namespace=namespace0,
                    name="c",
                    data_name=data_name,
                    header=header,
                )
            )
            out.append(
                f"{name}.set_construct("
                f"c, axes={self.get_data_axes(key)}, key={key!r}, copy=False)"
            )

        # Coordinate reference constructs
        for key, c in self.coordinate_references(todict=True).items():
            out.extend(
                c.creation_commands(
                    namespace=namespace0,
                    indent=0,
                    string=False,
                    name="c",
                    header=header,
                )
            )
            out.append(f"{name}.set_construct(c)")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    @_display_or_return
    def dump(
        self,
        display=True,
        _omit_properties=(),
        _prefix="",
        _title=None,
        _create_title=True,
        _level=0,
    ):
        """A full description of the domain construct.

        Returns a description of all properties, including those of
        metadata constructs and their components, and provides
        selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        indent = "    "
        indent0 = indent * _level

        if _create_title:
            if _title is None:
                ncvar = self.nc_get_variable(None)
                _title = self.identity(default=None)
                if ncvar is not None:
                    if _title is None:
                        _title = f"ncvar%{ncvar}"
                    else:
                        _title += f" (ncvar%{ncvar})"

                if _title is None:
                    _title = ""

                _title = f"{self.__class__.__name__}: {_title}"

            line = f"{indent0}{''.ljust(len(_title), '-')}"

            # Title
            string = [line, indent0 + _title, line]

            properties = super().dump(
                display=False,
                _create_title=False,
                _omit_properties=_omit_properties,
                _prefix=_prefix,
                _title=_title,
                _level=_level - 1,
            )
            string.append(properties)
            string.append("")
        else:
            string = []

        axis_to_name = self._unique_domain_axis_identities()

        construct_name = self._unique_construct_names()

        construct_data_axes = self.constructs.data_axes()

        # Domain axes
        axes = self._dump_axes(axis_to_name, display=False, _level=_level)
        if axes:
            string.append(axes)

        # Dimension coordinates
        dimension_coordinates = self.dimension_coordinates(todict=True)
        for cid, value in sorted(dimension_coordinates.items()):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _level=_level,
                    _title=f"Dimension coordinate: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        # Auxiliary coordinates
        auxiliary_coordinates = self.auxiliary_coordinates(todict=True)
        for cid, value in sorted(auxiliary_coordinates.items()):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _level=_level,
                    _title=f"Auxiliary coordinate: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        # Domain ancillaries
        for cid, value in sorted(self.domain_ancillaries(todict=True).items()):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _level=_level,
                    _title=f"Domain ancillary: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        # Coordinate references
        for cid, value in sorted(
            self.coordinate_references(todict=True).items()
        ):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _level=_level,
                    _title=f"Coordinate reference: {construct_name[cid]}",
                    _construct_names=construct_name,
                    _auxiliary_coordinates=tuple(auxiliary_coordinates),
                    _dimension_coordinates=tuple(dimension_coordinates),
                )
            )

        # Cell measures
        for cid, value in sorted(self.cell_measures(todict=True).items()):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _key=cid,
                    _level=_level,
                    _title=f"Cell measure: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        # Domain topologies
        for cid, value in sorted(self.domain_topologies(todict=True).items()):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _key=cid,
                    _level=_level,
                    _title=f"Domain topology: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        # Cell connectivities
        for cid, value in sorted(
            self.cell_connectivities(todict=True).items()
        ):
            string.append("")
            string.append(
                value.dump(
                    display=False,
                    _key=cid,
                    _level=_level,
                    _title=f"Cell connectivity: {construct_name[cid]}",
                    _axes=construct_data_axes[cid],
                    _axis_names=axis_to_name,
                )
            )

        string.append("")

        return "\n".join(string)

    def get_data_axes(self, *identity, default=ValueError(), **filter_kwargs):
        """Gets the keys of the axes spanned by the construct data.

        Specifically, returns the keys of the domain axis constructs
        spanned by the data of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `set_data_axes`

        :Parameters:

            identity, filter_kwargs: optional
                Select the unique construct returned by
                ``d.construct(*identity, **filter_kwargs)``. See
                `construct` for details.

                .. versionadded:: (cfdm) 1.10.0.0

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `tuple`
                The keys of the domain axis constructs spanned by the
                data.

        **Examples**

        >>> d = {{package}}.example_domain(0)
        >>> d.get_data_axes('latitude')
        ('domainaxis0',)
        >>> d.get_data_axes('time')
        ('domainaxis2',)
        >>> d.has_data_axes()
        True
        >>> d.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> d.has_data_axes()
        False
        >>> d.get_data_axes(default='no axes')
        'no axes'

        """
        key = self.construct(
            *identity, key=True, default=None, **filter_kwargs
        )
        if key is None:
            if default is None:
                return default

            return self._default(
                default, "Can't get axes for non-existent construct"
            )

        axes = super().get_data_axes(key, default=None)
        if axes is None:
            if default is None:
                return default

            return self._default(
                default, f"Construct {key!r} has not had axes set"
            )

        return axes

    def get_filenames(self, normalise=True):
        """Return the file names containing the metadata construct data.

        :Parameters:

            {{normalise: `bool`, optional}}

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `set`
                The file names in normalised, absolute form. If all of
                the data are in memory then an empty `set` is
                returned.

        **Examples**

        >>> d = {{package}}.example_domain(0)
        >>> {{package}}.write(d, 'temp_file.nc')
        >>> e = {{package}}.read('temp_file.nc', domain=True)[0]
        >>> e.get_filenames()
        {'temp_file.nc'}

        """
        out = set()

        for c in self.constructs.filter_by_data(todict=True).values():
            out.update(c.get_filenames(normalise=normalise))

        return out

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The ``cf_role`` property, preceded by ``'cf_role='``.
        * The ``long_name`` property, preceded by ``'long_name='``.
        * The netCDF variable name, preceded by ``'ncvar%'``.
        * The value of the *default* parameter.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of
                the default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> d = {{package}}.Domain()
        >>> d.set_properties({'foo': 'bar',
        ...                   'long_name': 'Domain for model'})
        >>> d.nc_set_variable('dom1')
        >>> d.identity()
        'long_name=Domain for model'
        >>> d.del_property('long_name')
        'long_name=Domain for model'
        >>> d.identity(default='no identity')
        'ncvar%dom1'
        >>> d.identity()
        'ncvar%dom1'
        >>> d.nc_del_variable()
        'dom1'
        >>> d.identity()
        ''
        >>> d.identity(default='no identity')
        'no identity'

        """
        for prop in ("cf_role", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self):
        """Return all possible identities.

        The identities comprise:

        * The ``cf_role`` property, preceded by ``'cf_role='``.
        * The ``long_name`` property, preceded by ``'long_name='``.
        * All other properties, preceded by the property name and a
          equals e.g. ``'foo=bar'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `identity`

        :Returns:

            `list`
                The identities.

        **Examples**

        >>> d = {{package}}.Domain()
        >>> d.set_properties({'foo': 'bar',
        ...                   'long_name': 'Domain for model'})
        >>> d.nc_set_variable('dom1')
        >>> d.identities()
        ['long_name=Domain for model', 'foo=bar', 'ncvar%dom1']

        """
        properties = self.properties()
        cf_role = properties.pop("cf_role", None)
        long_name = properties.pop("long_name", None)

        out = []

        if cf_role is not None:
            out.append(f"cf_role={cf_role}")

        if long_name is not None:
            out.append(f"long_name={long_name}")

        out += [
            f"{prop}={value}" for prop, value in sorted(properties.items())
        ]

        n = self.nc_get_variable(None)
        if n is not None:
            out.append(f"ncvar%{n}")

        return out

    @_inplace_enabled(default=False)
    def persist(self, inplace=False):
        """Persist data into memory.

        {{persist description}}

        **Performance**

        `persist` causes delayed operations to be computed.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `{{package}}.Data.persist`

        :Parameters:

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The domain construct with persisted metadata. If the
                operation was in-place then `None` is returned.

        """
        d = _inplace_enabled_define_and_cleanup(self)

        for c in d.constructs.filter_by_data(todict=True).values():
            c.persist(inplace=True)

        return d

    @_inplace_enabled(default=False)
    def uncompress(self, inplace=False):
        """Uncompress the domain construct.

        Compression saves space by identifying and removing unwanted
        missing data. Such compression techniques store the data more
        efficiently and result in no precision loss.  Whether or not
        the metadata constructs are compressed does not alter its
        functionality nor external appearance.

        Any compressed metadata constructs are uncompressed, and all
        other metadata constructs are unchanged.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `compress` TODO

        :Parameters:

            {{inplace: `bool`, optional}}

        :Returns:

            `Domain` or `None`
                The uncompressed domain construct, or `None` if the
                operation was in-place.

        **Examples**

        >>> e = d.uncompress()
        >>> e.equals(d)
        True

        """
        d = _inplace_enabled_define_and_cleanup(self)

        for c in d.constructs.filter_by_data(todict=True).values():
            c.uncompress(inplace=True)

        return d



from . import PropertiesDataBounds


class Coordinate(PropertiesDataBounds):
    """Mixin class for dimension or auxiliary coordinate constructs.

    .. versionadded:: (cfdm) 1.7.0

    """

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="c",
        data_name="data",
        bounds_name="b",
        interior_ring_name="i",
        header=True,
    ):
        """Return the commands that would create the construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{data_name: `str`, optional}}

            bounds_name: `str`, optional
                The name of the construct's `Bounds` instance created
                by the returned commands.

                *Parameter example:*
                  ``name='bounds1'``

            interior_ring_name: `str`, optional
                The name of the construct's `InteriorRing` instance
                created by the returned commands.

                *Parameter example:*
                  ``name='ir1'``

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.{{class}}(
        ...     properties={'units': 'degrees_east',
        ...                 'standard_name': 'longitude'}
        ... )
        >>> x.set_data([22.5, 67.5, 112.5])
        >>> b = {{package}}.Bounds()
        >>> b.set_data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0]])
        >>> x.set_bounds(b)
        >>> print(x.creation_commands(header=False))
        c = {{package}}.{{class}}()
        c.set_properties({'units': 'degrees_east', 'standard_name': 'longitude'})
        data = {{package}}.Data([22.5, 67.5, 112.5], units='degrees_east', dtype='f8')
        c.set_data(data)
        b = {{package}}.Bounds()
        data = {{package}}.Data([[0.0, 45.0], [45.0, 90.0], [90.0, 135.0]], units='degrees_east', dtype='f8')
        b.set_data(data)
        c.set_bounds(b)

        """
        out = super().creation_commands(
            representative_data=representative_data,
            namespace=namespace,
            indent=0,
            string=False,
            name="c",
            data_name="data",
            bounds_name="b",
            interior_ring_name="i",
            header=header,
            _coordinate=True,
        )

        try:
            nc = self.nc_get_node_coordinate_variable(None)
        except AttributeError:
            pass
        else:
            if nc is not None:
                out.append(f"{name}.nc_set_node_coordinate_variable({nc!r})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out



from . import Constructs, abstract, mixin


class Domain(mixin.FieldDomain, abstract.Properties):
    """A domain construct of the CF data model.

    The domain represents a set of discrete "locations" in what
    generally would be a multi-dimensional space, either in the real
    world or in a model's simulated world. The data array elements of
    a field construct correspond to individual location of a domain.

    The domain construct is defined collectively by the following
    constructs of the CF data model: domain axis, dimension
    coordinate, auxiliary coordinate, cell measure, coordinate
    reference, and domain ancillary constructs; as well as properties
    to describe the domain.

    .. versionadded:: (cfdm) 1.7.0

    """

    # Define the base of the identity keys for each construct type
    _construct_key_base = {
        "auxiliary_coordinate": "auxiliarycoordinate",
        "cell_connectivity": "cellconnectivity",
        "cell_measure": "cellmeasure",
        "coordinate_reference": "coordinatereference",
        "dimension_coordinate": "dimensioncoordinate",
        "domain_ancillary": "domainancillary",
        "domain_axis": "domainaxis",
        "domain_topology": "domaintopology",
    }

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance

    def __init__(
        self, properties=None, source=None, copy=True, _use_data=True
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

            *Parameter example:*
               ``properties={'long_name': 'Domain for model'}``

            {{init source: optional}}

                A new domain may also be instantiated with the
                `fromconstructs` class method.

            {{init copy: `bool`, optional}}

        """
        super().__init__(properties=properties, source=source, copy=copy)

        if source is not None:
            try:
                constructs = source.constructs
            except AttributeError:
                constructs = self._Constructs(**self._construct_key_base)
                copy = False
                _use_data = True
            else:
                constructs = constructs._view(
                    ignore=("cell_method", "field_ancillary")
                )
        else:
            constructs = self._Constructs(**self._construct_key_base)
            copy = False
            _use_data = True

        if copy or not _use_data:
            constructs = constructs.copy(data=_use_data)

        self._set_component("constructs", constructs, copy=False)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.9.0.0

        :Returns:
            `str`
                The construct type.

        **Examples**

        >>> d = {{package}}.{{class}}()
        >>> d.construct_type
        'domain'

        """
        return "domain"

    @property
    def constructs(self):
        """Return the metadata constructs.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `Constructs`
                The constructs.

        **Examples**

        >>> d = {{package}}.example_field(0)
        >>> print(d.constructs)
        Constructs:
        {'cellmethod0': <CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}

        """
        return self._get_component("constructs")

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, data=True):
        """Return a deep copy.

        ``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `fromconstructs`

        :Parameters:

            {{data: `bool`, optional}}

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> d = {{package}}.{{class}}()
        >>> e = d.copy()

        """
        return type(self)(source=self, copy=True, _use_data=data)

    def del_construct(self, key, default=ValueError()):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it
        can't be spanned by any data arrays of the metadata
        constructs. However, a domain ancillary construct may be
        removed even if it is referenced by coordinate reference
        construct.

        .. versionadded:: (cfdm) 1.9.0.0

        .. seealso:: `get_construct`, `constructs`, `has_construct`,
                     `set_construct`

        :Parameters:

            key: `str`
                The construct identifier of the metadata construct to
                be removed.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

        :Returns:

                The removed metadata construct.

        **Examples**

        >>> f.del_construct('auxiliarycoordinate2')
        <{{repr}}AuxiliaryCoordinate: latitude(111, 106) degrees_north>
        >>> f.del_construct('auxiliarycoordinate2')
        ValueError: Can't get remove non-existent construct
        >>> f.del_construct('auxiliarycoordinate2', default=False)
        False

        """
        return self.constructs._del_construct(key, default=default)

    @classmethod
    def fromconstructs(cls, constructs, copy=False):
        """Return a new domain containing the given metadata constructs.

        The new domain acts as a view to the given constructs,
        i.e. changes to the domain, such as the addition or removal of
        a construct, will also affect the input `Constructs` instance.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            constructs: `Constructs`
                The constructs from which to create the new
                domain. Cell method and field ancillary constructs are
                ignored.

            copy: `bool`, optional
                If True then deep copy the metadata constructs prior
                to initialisation. By default the metadata constructs
                are not copied. Note that even when *copy* is True,
                the input `Constructs` container is not copied.

        :Returns:

            `{{class}}`
                The domain created from a view of the constructs.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> d = {{package}}.{{class}}.fromconstructs(f.constructs)
        >>> d
        <Domain: {1, 5, 8}>
        >>> d = {{package}}.{{class}}.fromconstructs(f.constructs.copy())

        """
        domain = cls()
        domain._set_component(
            "constructs",
            constructs._view(ignore=("cell_method", "field_ancillary")),
            copy=copy,
        )

        domain.constructs._field_data_axes = None

        return domain



"""A data indexer that applies netCDF masking and unpacking.

Portions of this code were adapted from the `netCDF4` Python library,
which carries the following MIT License:

Copyright 2008 Jeffrey Whitaker

https://opensource.org/license/mit

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

"""

import logging
from math import prod
from numbers import Integral

import numpy as np
from dask.array.slicing import normalize_index
from netCDF4 import chartostring, default_fillvals
from netCDF4.utils import _safecast

logger = logging.getLogger(__name__)


class netcdf_indexer:
    """A data indexer that also applies netCDF masking and unpacking.

    Indexing may be orthogonal or non-orthogonal. Orthogonal indexing
    means that the index for each dimension is applied independently,
    regardless of how that index was defined. For instance, the
    indices ``[[0, 1], [1, 3], 0]`` and ``[:2, 1:4:2, 0]`` will give
    identical results. Orthogonal indexing is different to the
    indexing behaviour of `numpy`. Non-orthogonal indexing means that
    normal `numpy` indexing rules are applied.

    In addition, string and character variables are always converted
    to unicode arrays, the latter with the last dimension
    concatenated.

    Masking and unpacking operations, either or both may be disabled
    via initialisation options, are defined by the conventions for
    netCDF attributes, which are either provided as part of the input
    *variable* object, or given with the input *attributes* parameter.

    The relevant netCDF attributes that are considered are:

      * For masking: ``missing_value``, ``valid_max``, ``valid_min``,
                     ``valid_range``, ``_FillValue``, and
                     ``_Unsigned``. Note that if ``_FillValue`` is not
                     present then the netCDF default value for the
                     appropriate data type will be assumed, as defined
                     by `netCDF4.default_fillvals`.

      * For unpacking: ``add_offset``, ``scale_factor``, and
                       ``_Unsigned``

    .. versionadded:: (cfdm) 1.11.2.0

    **Examples**

    >>> import netCDF4
    >>> nc = netCDF4.Dataset('file.nc', 'r')
    >>> x = cfdm.netcdf_indexer(nc.variables['x'])
    >>> x.shape
    (12, 64, 128)
    >>> print(x[0, 0:4, 0:3])
    [[236.5, 236.2, 236.0],
     [240.9, --   , 239.6],
     [243.4, 242.4, 241.3],
     [243.1, 241.7, 240.4]]

    >>> import h5netcdf
    >>> h5 = h5netcdf.File('file.nc', 'r')
    >>> x = cfdm.netcdf_indexer(h5.variables['x'])
    >>> x.shape
    (12, 64, 128)
    >>> print(x[0, 0:4, 0:3])
    [[236.5, 236.2, 236.0],
     [240.9, --   , 239.6],
     [243.4, 242.4, 241.3],
     [243.1, 241.7, 240.4]]

    >>> import numpy as np
    >>> n = np.arange(7)
    >>> x = cfdm.netcdf_indexer(n)
    >>> x.shape
    (7,)
    >>> print(x[...])
    [0 1 2 3 4 5 6]
    >>> x = cfdm.netcdf_indexer(n, attributes={'_FillValue': 4})
    >>> print(x[...])
    [0 1 2 3 -- 5 6]
    >>> x = cfdm.netcdf_indexer(n, mask=False, attributes={'_FillValue': 4})
    >>> print(x[...])
    [0 1 2 3 4 5 6]

    """

    def __init__(
        self,
        variable,
        mask=True,
        unpack=True,
        always_masked_array=False,
        orthogonal_indexing=False,
        attributes=None,
        copy=False,
    ):
        """**Initialisation**

        :Parameters:

            variable:
                The variable to be indexed. May be any variable that
                has the same API as one of `numpy.ndarray`,
                `netCDF4.Variable`, or `h5py.Variable` (which includes
                `h5netcdf.Variable`). Any masking and unpacking that
                could be applied by *variable* itself (e.g. by a
                `netCDF4.Variable` instance) is disabled, ensuring
                that any masking and unpacking is always done by the
                `netcdf_indexer` instance.

            mask: `bool`, optional
                If True, the default, then an array returned by
                indexing is automatically masked. Masking is
                determined by the netCDF conventions for the following
                attributes: ``_FillValue``, ``missing_value``,
                ``_Unsigned``, ``valid_max``, ``valid_min``, and
                ``valid_range``.

            unpack: `bool`, optional
                If True, the default, then an array returned by
                indexing is automatically unpacked. Unpacking is
                determined by the netCDF conventions for the following
                attributes: ``add_offset``, ``scale_factor``, and
                ``_Unsigned``.

            always_masked_array: `bool`, optional
                If False, the default, then an array returned by
                indexing which has no missing values is created as a
                regular `numpy` array. If True then an array returned
                by indexing is always a masked `numpy` array, even if
                there are no missing values.

            orthogonal_indexing: `bool`, optional
                If True then indexing is orthogonal, meaning that the
                index for each dimension is applied independently,
                regardless of how that index was defined. For
                instance, the indices ``[[0, 1], [1, 3], 0]`` and
                ``[:2, 1:4:2, 0]`` will give identical results. This
                behaviour is different to that of `numpy`. If False,
                the default, then normal `numpy` indexing rules are
                applied.

            attributes: `dict`, optional
                Provide netCDF attributes for the *variable* as a
                dictionary of key/value pairs. Only the attributes
                relevant to masking and unpacking are considered, with
                all other attributes being ignored. If *attributes* is
                `None`, the default, then the netCDF attributes stored
                by *variable* (if any) are used. If *attributes* is
                not `None`, then any netCDF attributes stored by
                *variable* are ignored.

            copy: `bool`, optional
                If True then return a `numpy` array that is not a view
                of part of the original data, i.e. in-place
                changes to the returned subspace will not affect the
                original *variable*. This is done by returning an
                in-memory copy the subspace. If False, the default, no
                in-memory copy is made, and then whether or not
                in-place changes to the returned subspace affect
                *variable* will depend on how subspacing is
                implemented by *variable*.

        """
        self.variable = variable
        self.mask = bool(mask)
        self.unpack = bool(unpack)
        self.always_masked_array = bool(always_masked_array)
        self._attributes = attributes
        self._copy = bool(copy)
        self._orthogonal_indexing = bool(orthogonal_indexing)

    def __getitem__(self, index):
        """Return a subspace of the variable as a `numpy` array.

        n.__getitem__(index) <==> v[index]

        If `__orthogonal_indexing__` is True then indexing is
        orthogonal.  If `__orthogonal_indexing__` is False then normal
        `numpy` indexing rules are applied.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        variable = self.variable
        unpack = self.unpack
        attributes = self.attributes()
        dtype = variable.dtype

        # Prevent a netCDF4 variable from doing its own masking and
        # unpacking during the indexing
        netCDF4_scale = False
        netCDF4_mask = False
        try:
            netCDF4_scale = variable.scale
        except AttributeError:
            # Not a netCDF4 variable
            pass
        else:
            netCDF4_mask = variable.mask
            variable.set_auto_maskandscale(False)

        # ------------------------------------------------------------
        # Index the variable
        # ------------------------------------------------------------
        try:
            data = self._index(index)
        except (IndexError, AttributeError):
            # Assume we are here because we have one or more
            # np.newaxis values in 'index', and the variable doesn't
            # support that type of indexing. It is known that
            # `netCDF4` raises an IndexError and h5netcdf raises an
            # AttributeError.

            # Subspace the variable with the np.newaxis elements
            # removed
            newaxis = np.newaxis
            index1 = [i for i in index if i is not newaxis]
            data = self._index(tuple(index1))

            # Now subspace the result (which we're assuming is
            # something that likes np.newaxis indices) with the
            # np.newaxis elements reinstated.
            index2 = [i if i is newaxis else slice(None) for i in index]
            data = self._index(tuple(index2), data=data)

            # E.g.     index : (1, np.newaxis, slice(1, 5))
            #      =>  index1: (1, slice(1, 5))
            #      and index2: (slice(None), np.newaxis, slice(None))
        except ValueError:
            # Something went wrong, which is indicative of the
            # variable not supporting the appropriate slicing method
            # (e.g. `h5netcdf` might have returned "ValueError: Step
            # must be >= 1 (got -2)"). Therefore we'll just get the
            # entire array as a numpy array, and then try indexing
            # that.
            data = self._index(Ellipsis)
            data = self._index(index, data=data)

        # Reset a netCDF4 variable's scale and mask behaviour
        if netCDF4_scale:
            variable.set_auto_scale(True)

        if netCDF4_mask:
            variable.set_auto_mask(True)

        # Convert str, char, and object data to byte strings
        if isinstance(data, str):
            data = np.array(data, dtype="S")
        elif data.dtype.kind in "OSU":
            kind = data.dtype.kind
            if kind == "S":
                data = chartostring(data)

            # Assume that object arrays are arrays of strings
            data = data.astype("S", copy=False)
            if kind == "O":
                dtype = data.dtype

        if dtype is str:
            dtype = data.dtype

        dtype_unsigned_int = None
        if unpack:
            is_unsigned_int = attributes.get("_Unsigned") in ("true", "True")
            if is_unsigned_int:
                data_dtype = data.dtype
                dtype_unsigned_int = (
                    f"{data_dtype.byteorder}u{data_dtype.itemsize}"
                )
                data = data.view(dtype_unsigned_int)

        # ------------------------------------------------------------
        # Mask the data
        # ------------------------------------------------------------
        if self.mask:
            data = self._mask(data, dtype, attributes, dtype_unsigned_int)

        # ------------------------------------------------------------
        # Unpack the data
        # ------------------------------------------------------------
        if unpack:
            data = self._unpack(data, attributes)

        # Make sure all strings are unicode
        if data.dtype.kind == "S":
            data = data.astype("U", copy=False)

        # ------------------------------------------------------------
        # Copy the data
        # ------------------------------------------------------------
        if self._copy:
            data = data.copy()

        return data

    @property
    def __orthogonal_indexing__(self):
        """Flag to indicate whether indexing is orthogonal.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return self._orthogonal_indexing

    def _check_safecast(self, attr, dtype, attributes):
        """Check an attribute's data type.

        Checks to see that variable attribute exists and can be safely
        cast to variable's data type.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameter:

            attr: `str`
                The name of the attribute.

            dtype: `numpy.dtype`
                The variable data type.

            attributes: `dict`
                The variable attributes.

        :Returns:

            `bool`, value
                Whether or not the attribute data type is consistent
                with the variable data type, and the attribute value.

        """
        if attr in attributes:
            attvalue = attributes[attr]
            att = np.array(attvalue)
        else:
            return False, None

        try:
            atta = np.array(att, dtype)
        except ValueError:
            safe = False
        else:
            safe = _safecast(att, atta)

        if not safe:
            logger.info(
                f"Mask attribute {attr!r} not used since it can't "
                f"be safely cast to variable data type {dtype!r}"
            )  # pragma: no cover

        return safe, attvalue

    def _default_FillValue(self, dtype):
        """Return the default ``_FillValue`` for the given data type.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `netCDF4.default_fillvals`

        :Parameter:

            dtype: `numpy.dtype`
                The data type.

        :Returns:

                The default ``_FillValue``.

        """
        if dtype.kind in "OS":
            return default_fillvals["S1"]

        return default_fillvals[dtype.str[1:]]

    def _index(self, index, data=None):
        """Get a subspace of the variable.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `__getitem__`

        :Parameter:

            index:
                The indices that define the subspace.

            data: array_like, optional
                The data to be indexed. If `None` (the default) then
                the data given by the `variable` attribute will be
                used.

        :Returns:

            `numpy.ndarray`
                The subspace of the variable.

        """
        if data is None:
            data = self.variable

        if index is Ellipsis:
            return data[...]

        index = normalize_index(index, data.shape)

        # Find the positions of any list/1-d array indices (which by
        # now will contain only integers)
        axes_with_list_indices = [
            n
            for n, i in enumerate(index)
            if isinstance(i, list) or getattr(i, "shape", False)
        ]

        data_orthogonal_indexing = getattr(
            data, "__orthogonal_indexing__", False
        )
        if not self.__orthogonal_indexing__:
            # --------------------------------------------------------
            # Do non-orthogonal indexing
            # --------------------------------------------------------
            if data_orthogonal_indexing and len(axes_with_list_indices) > 1:
                raise IndexError(
                    "Can't non-orthogonally index "
                    f"{data.__class__.__name__} with index {index!r}"
                )

            return data[index]

        # ------------------------------------------------------------
        # Still here? Then do orthogonal indexing.
        # ------------------------------------------------------------

        # Create an index that replaces integers with size 1 slices,
        # so that their axes are not dropped yet (they will be dropped
        # later).
        index0 = [
            slice(i, i + 1) if isinstance(i, Integral) else i for i in index
        ]

        if data_orthogonal_indexing or len(axes_with_list_indices) <= 1:
            # There is at most one list/1-d array index, and/or the
            # variable natively supports orthogonal indexing.
            #
            # Note: `netCDF4.Variable` natively supports orthogonal
            #       indexing; but `h5netcdf.File`, `h5py.File`, and
            #       `numpy.ndarray`, do not.
            data = data[tuple(index0)]
        else:
            # There are two or more list/1-d array indices, and the
            # variable does not natively support orthogonal indexing
            # => emulate orthogonal indexing with a sequence of
            # independent subspaces, one for each list/1-d array
            # index.

            # 1) Apply the slice indices at the same time as the
            #    list/1-d array index that gives the smallest result.

            # Create an index that replaces each list/1-d array with
            # slice(None)
            index1 = [
                i if isinstance(i, slice) else slice(None) for i in index0
            ]

            # Find the position of the list/1-d array index that gives
            # the smallest result, and apply the subspace of slices
            # and the chosen list/1-d array index. This will give the
            # smallest high-water memory mark of the whole operation.
            shape1 = self.index_shape(index1, data.shape)
            size1 = prod(shape1)
            sizes = [
                size1 * (len(index[i]) // shape1[i])
                for i in axes_with_list_indices
            ]
            n = axes_with_list_indices.pop(np.argmin(sizes))
            index1[n] = index[n]
            data = data[tuple(index1)]

            # 2) Apply the rest of the list/1-d array indices, in the
            #    order that gives the smallest result after each step.
            ndim = data.ndim
            while axes_with_list_indices:
                shape1 = data.shape
                size1 = data.size
                sizes = [
                    len(index[i]) * size1 // shape1[i]
                    for i in axes_with_list_indices
                ]
                n = axes_with_list_indices.pop(np.argmin(sizes))
                index2 = [slice(None)] * ndim
                index2[n] = index[n]
                data = data[tuple(index2)]

        # Apply any integer indices that will drop axes
        index3 = [0 if isinstance(i, Integral) else slice(None) for i in index]
        if index3:
            data = data[tuple(index3)]

        return np.asanyarray(data)

    def _mask(self, data, dtype, attributes, dtype_unsigned_int):
        """Mask the data.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameter:

            data: `numpy.ndarray`
                The unmasked and (possibly) packed data.

            dtype: `numpy.dtype`
                The data type of the variable (which may be different
                to that of *data*).

            attributes: `dict`
                The variable attributes.

            dtype_unsigned_int: `dtype` or `None`
                The data type when the data have been cast to unsigned
                integers, otherwise `None`.

        :Returns:

            `numpy.ndarray`
                The masked data.

        """
        # The Boolean mask accounting for all methods of specification
        totalmask = None
        # The fill value for the returned numpy array
        fill_value = None

        safe_missval, missing_value = self._check_safecast(
            "missing_value", dtype, attributes
        )
        if safe_missval:
            # --------------------------------------------------------
            # Create mask from missing_value
            # --------------------------------------------------------
            mval = np.array(missing_value, dtype)
            if dtype_unsigned_int is not None:
                mval = mval.view(dtype_unsigned_int)

            if not mval.ndim:
                mval = (mval,)

            for m in mval:
                try:
                    mvalisnan = np.isnan(m)
                except TypeError:
                    # isnan fails on some dtypes
                    mvalisnan = False

                if mvalisnan:
                    mask = np.isnan(data)
                else:
                    mask = data == m

                if mask.any():
                    if totalmask is None:
                        totalmask = mask
                    else:
                        totalmask += mask

            if totalmask is not None:
                fill_value = mval[0]

        # Set mask=True for data == fill value
        safe_fillval, _FillValue = self._check_safecast(
            "_FillValue", dtype, attributes
        )
        if not safe_fillval:
            _FillValue = self._default_FillValue(dtype)
            safe_fillval = True

        if safe_fillval:
            # --------------------------------------------------------
            # Create mask from _FillValue
            # --------------------------------------------------------
            fval = np.array(_FillValue, dtype)
            if dtype_unsigned_int is not None:
                fval = fval.view(dtype_unsigned_int)

            if fval.ndim == 1:
                # _FillValue must be a scalar
                fval = fval[0]

            try:
                fvalisnan = np.isnan(fval)
            except Exception:
                # isnan fails on some dtypes
                fvalisnan = False

            if fvalisnan:
                mask = np.isnan(data)
            else:
                mask = data == fval

            if mask.any():
                if fill_value is None:
                    fill_value = fval

                if totalmask is None:
                    totalmask = mask
                else:
                    totalmask += mask

        # Set mask=True for data outside [valid_min, valid_max]
        #
        # If valid_range exists use that, otherwise look for
        # valid_min, valid_max. No special treatment of byte data as
        # described in the netCDF documentation.
        validmin = None
        validmax = None
        safe_validrange, valid_range = self._check_safecast(
            "valid_range", dtype, attributes
        )
        safe_validmin, valid_min = self._check_safecast(
            "valid_min", dtype, attributes
        )
        safe_validmax, valid_max = self._check_safecast(
            "valid_max", dtype, attributes
        )
        if safe_validrange and valid_range.size == 2:
            validmin = np.array(valid_range[0], dtype)
            validmax = np.array(valid_range[1], dtype)
        else:
            if safe_validmin:
                validmin = np.array(valid_min, dtype)

            if safe_validmax:
                validmax = np.array(valid_max, dtype)

        if dtype_unsigned_int is not None:
            if validmin is not None:
                validmin = validmin.view(dtype_unsigned_int)

            if validmax is not None:
                validmax = validmax.view(dtype_unsigned_int)

        if dtype.kind != "S":
            # --------------------------------------------------------
            # Create mask from valid_min. valid_max, valid_range
            # --------------------------------------------------------
            # Don't set validmin/validmax mask for character data
            #
            # Setting valid_min/valid_max to the _FillVaue is too
            # surprising for many users (despite the netcdf docs
            # attribute best practices suggesting clients should do
            # this).
            if validmin is not None:
                if validmin.ndim == 1:
                    # valid min must be a scalar
                    validmin = validmin[0]

                mask = data < validmin
                if totalmask is None:
                    totalmask = mask
                else:
                    totalmask += mask

            if validmax is not None:
                if validmax.ndim == 1:
                    # valid max must be a scalar
                    validmax = validmax[0]

                mask = data > validmax
                if totalmask is None:
                    totalmask = mask
                else:
                    totalmask += mask

        # ------------------------------------------------------------
        # Mask the data
        # ------------------------------------------------------------
        if totalmask is not None and totalmask.any():
            data = np.ma.masked_array(
                data, mask=totalmask, fill_value=fill_value, copy=False
            )
            if not data.ndim:
                # Return a scalar numpy masked constant not a 0-d
                # masked array, so that data == np.ma.masked.
                data = data[()]
        elif np.ma.isMA(data):
            if not (self.always_masked_array or np.ma.is_masked(data)):
                # Return a non-masked array
                data = np.array(data)
        elif self.always_masked_array:
            # Return a masked array
            data = np.ma.masked_array(data)

        return data

    def _unpack(self, data, attributes):
        """Unpack the data.

        If both the ``add_offset`` and ``scale_factor`` attributes
        have not been set then no unpacking is done and the data is
        returned unchanged.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameter:

            data: `numpy.ndarray`
                The masked and (possibly) packed data.

            attributes: `dict`
                The variable attributes.

        :Returns:

            `numpy.ndarray`
                The unpacked data.

        """
        scale_factor = attributes.get("scale_factor")
        add_offset = attributes.get("add_offset")

        try:
            if scale_factor is not None:
                scale_factor = np.array(scale_factor)
                if scale_factor.ndim == 1:
                    # scale_factor must be a scalar
                    scale_factor = scale_factor[0]

                float(scale_factor)
        except ValueError:
            logging.warn(
                "No unpacking done: 'scale_factor' attribute "
                f"{scale_factor!r} can't be converted to a float"
            )  # pragma: no cover
            return data

        try:
            if add_offset is not None:
                add_offset = np.array(add_offset)
                if add_offset.ndim == 1:
                    # add_offset must be a scalar
                    add_offset = add_offset[0]

                float(add_offset)
        except ValueError:
            logging.warn(
                "No unpacking done: 'add_offset' attribute "
                f"{add_offset!r} can't be converted to a float"
            )  # pragma: no cover
            return data

        if scale_factor is not None:
            if add_offset is not None:
                # scale_factor and add_offset
                if add_offset != 0.0 or scale_factor != 1.0:
                    data = data * scale_factor + add_offset
                    self._copy = False
                else:
                    data = data.astype(np.array(scale_factor).dtype)
            else:
                # scale_factor with no add_offset
                if scale_factor != 1.0:
                    data = data * scale_factor
                    self._copy = False
                else:
                    data = data.astype(scale_factor.dtype)
        elif add_offset is not None:
            # add_offset with no scale_factor
            if add_offset != 0.0:
                data = data + add_offset
                self._copy = False
            else:
                data = data.astype(np.array(add_offset).dtype)

        return data

    def _size_1_axis(self):
        """Find the position of a unique size 1 index.

        .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `int` or `None`
                The position of the unique size 1 index, or `None` if
                there are zero or at least two of them.

        """
        shape = self.shape
        if shape.count(1):
            return shape.index(1)

        return

    @property
    def dtype(self):
        """The data type of the array elements.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return self.variable.dtype

    @property
    def ndim(self):
        """Number of dimensions in the data array.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return self.variable.ndim

    @property
    def shape(self):
        """Tuple of the data dimension sizes.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return self.variable.shape

    @property
    def size(self):
        """Number of elements in the data array.

        .. versionadded:: (cfdm) 1.11.2.0

        """
        return self.variable.size

    def attributes(self):
        """Return the netCDF attributes for the data.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `dict`
                The attributes.

        **Examples**

        >>> n.attributes()
        {'standard_name': 'air_temperature',
         'missing_value': -999.0}

        """
        _attributes = self._attributes
        if _attributes is not None:
            return _attributes.copy()

        variable = self.variable
        try:
            # h5py API
            attrs = dict(variable.attrs)
        except AttributeError:
            try:
                # netCDF4 API
                attrs = {
                    attr: variable.getncattr(attr)
                    for attr in variable.ncattrs()
                }
            except AttributeError:
                # numpy API
                attrs = {}

        self._attributes = attrs
        return attrs

    @classmethod
    def index_shape(cls, index, shape):
        """Return the shape of the array subspace implied by indices.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            index: `tuple`
                The indices to be applied to an array with shape
                *shape*.

            shape: sequence of `int`
                The shape of the array to be subspaced.

        :Returns:

            `list`
                The shape of the subspace defined by the *index*.

        **Examples**

        >>> import numpy as np
        >>> n.index_shape((slice(2, 5), [4]), (10, 20))
        [3, 1]
        >>> n.index_shape((slice(2, 5), 4), (10, 20))
        [3]
        >>> n.index_shape(([2, 3, 4], np.arange(1, 6)), (10, 20))
        [3, 5]

        >>> n.index_shape((slice(None), [True, False, True]), (10, 3))
        [10, 2]

        >>> index0 = np.arange(5)
        >>> index0 = index0[index0 < 3]
        >>> n.index_shape((index0, []), (10, 20))
        [3, 0]

        >>> n.index_shape((slice(1, 5, 3), [3]), (10, 20))
        [2, 1]
        >>> n.index_shape((slice(5, 1, -2), 3), (10, 20))
        [2]
        >>> n.index_shape((slice(5, 1, 3), [3]), (10, 20))
        [0, 1]
        >>> n.index_shape((slice(1, 5, -3), 3), (10, 20))
        [0]

        """
        implied_shape = []
        for ind, full_size in zip(index, shape):
            if isinstance(ind, slice):
                start, stop, step = ind.indices(full_size)
                if (stop - start) * step < 0:
                    # E.g. 5:1:3 or 1:5:-3
                    size = 0
                else:
                    size = abs((stop - start) / step)
                    int_size = round(size)
                    if size > int_size:
                        size = int_size + 1
                    else:
                        size = int_size
            elif isinstance(ind, np.ndarray):
                if ind.dtype == bool:
                    # Size is the number of True values in the array
                    size = int(ind.sum())
                else:
                    size = ind.size

                if not ind.ndim:
                    # Scalar array
                    continue
            elif isinstance(ind, list):
                if not ind:
                    size = 0
                else:
                    i = ind[0]
                    if isinstance(i, bool):
                        # List of bool: Size is the number of True
                        # values in the list
                        size = sum(ind)
                    else:
                        # List of int
                        size = len(ind)
            else:
                # Index is Integral
                continue

            implied_shape.append(size)

        return implied_shape



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

warnings = False

# Set up temporary files
n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_read_write.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
[tmpfile1] = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class UGRIDTest(unittest.TestCase):
    """Test UGRID field constructs."""

    filename1 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "ugrid_1.nc"
    )

    filename2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "ugrid_2.nc"
    )

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows: cfdm.LOG_LEVEL('DEBUG')
        #
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_UGRID_read(self):
        """Test reading of UGRID files."""
        f1 = cfdm.read(self.filename1)

        self.assertEqual(len(f1), 3)
        for g in f1:
            self.assertEqual(len(g.domain_topologies()), 1)
            self.assertEqual(len(g.auxiliary_coordinates()), 2)
            self.assertEqual(len(g.dimension_coordinates()), 1)

            for aux in g.auxiliary_coordinates().values():
                self.assertTrue(aux.has_data())

            if g.domain_topology().get_cell() == "face":
                self.assertEqual(len(g.cell_connectivities()), 1)
                self.assertEqual(
                    g.cell_connectivity().get_connectivity(), "edge"
                )

        # Check that all fields have the same mesh id
        mesh_ids1 = set(g.get_mesh_id() for g in f1)
        self.assertEqual(len(mesh_ids1), 1)

        f2 = cfdm.read(self.filename2)
        self.assertEqual(len(f2), 3)
        for g in f2:
            self.assertEqual(len(g.domain_topologies()), 1)
            self.assertEqual(len(g.auxiliary_coordinates()), 2)
            self.assertEqual(len(g.dimension_coordinates()), 1)

            cell = g.domain_topology().get_cell()
            if cell in ("edge", "face"):
                for aux in g.auxiliary_coordinates().values():
                    self.assertFalse(aux.has_data())

            if cell == "face":
                self.assertEqual(len(g.cell_connectivities()), 1)
                self.assertEqual(
                    g.cell_connectivity().get_connectivity(), "edge"
                )

        # Check that all fields have the same mesh id
        mesh_ids2 = set(g.get_mesh_id() for g in f2)
        self.assertEqual(len(mesh_ids2), 1)

        # Check that the different files have different mesh ids
        self.assertNotEqual(mesh_ids1, mesh_ids2)

    def test_UGRID_data(self):
        """Test reading of UGRID data."""
        node1, face1, edge1 = cfdm.read(self.filename1)
        node2, face2, edge2 = cfdm.read(self.filename2)

        # Domain topology arrays
        domain_topology1 = face1.domain_topology()
        self.assertTrue(
            (
                domain_topology1.array
                == np.array([[2, 3, 1, 0], [4, 5, 3, 2], [1, 3, 6, -99]])
            ).all()
        )
        self.assertTrue(domain_topology1.equals(face2.domain_topology()))

        domain_topology1 = edge1.domain_topology()
        self.assertTrue(
            (
                domain_topology1.array
                == np.array(
                    [
                        [1, 6],
                        [3, 6],
                        [3, 1],
                        [0, 1],
                        [2, 0],
                        [2, 3],
                        [2, 4],
                        [5, 4],
                        [3, 5],
                    ]
                )
            ).all()
        )
        self.assertTrue(domain_topology1.equals(edge2.domain_topology()))

        # Cell connectivity arrays
        cell_connectivity1 = face1.cell_connectivity()
        self.assertTrue(
            (
                cell_connectivity1.array
                == np.array(
                    [
                        [0, 1, 2, -99, -99],
                        [1, 0, -99, -99, -99],
                        [2, 0, -99, -99, -99],
                    ]
                )
            ).all()
        )
        self.assertTrue(cell_connectivity1.equals(face2.cell_connectivity()))

    def test_read_UGRID_domain(self):
        """Test reading of UGRID files into domains."""
        d1 = cfdm.read(self.filename1, domain=True)

        self.assertEqual(len(d1), 3)
        for g in d1:
            self.assertIsInstance(g, cfdm.Domain)
            self.assertEqual(len(g.domain_topologies()), 1)
            self.assertEqual(len(g.auxiliary_coordinates()), 2)
            self.assertEqual(len(g.dimension_coordinates()), 0)

            for aux in g.auxiliary_coordinates().values():
                self.assertTrue(aux.has_data())

            if g.domain_topology().get_cell() == "face":
                self.assertEqual(len(g.cell_connectivities()), 1)
                self.assertEqual(
                    g.cell_connectivity().get_connectivity(), "edge"
                )

        # Check that all domains have the same mesh id
        mesh_ids1 = set(g.get_mesh_id() for g in d1)
        self.assertEqual(len(mesh_ids1), 1)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)
